// (c) 2002 by Jrn Loviscach

#include <fstream>
#include "bezierpath.h"
#include "bezierpathcollection.h"

#include "adapt.h"

BezierPathCollection::BezierPathCollection()
: hasPath(false)
{}

BezierPathCollection::~BezierPathCollection()
{
	if(hasPath)
	{
		hasPath = false;
		for(int i=0; i<pathColl.size(); ++i)
		{
			DELETE(pathColl[i]);
		}
	}
}

void BezierPathCollection::Trace(const PixelPathCollection& pixPathColl, double fitPrec)
{
	if(hasPath)
	{
		hasPath = false;
		for(int i=pathColl.size()-1; i>=0; --i)
		{
			DELETE(pathColl[i]);
			pathColl.pop_back();
		}
	}
	assert(pathColl.empty());

	width_ = pixPathColl.GetWidth();
	height_ = pixPathColl.GetHeight();

	for(int i=0; i<pixPathColl.GetLength(); ++i)
	{
		BezierPath* bp = NEW(BezierPath);
		bp->Init(*pixPathColl.GetPath(i));
		bp->Fit(fitPrec);
		pathColl.push_back(bp);
	}

	hasPath = true;
}

void BezierPathCollection::WriteEPSFile(const std::string filename) const
{
	assert(hasPath);

	std::ofstream f(filename.c_str()); // text mode implied
	f.exceptions(std::ios_base::badbit | std::ios_base::failbit);
	f.precision(2);
	f.setf(std::ios_base::fixed);
	f << "%!PS-Adobe-3.0 EPSF-3.0" << std::endl
		<< "%%BoundingBox: 0 0 "
		<< width_ << " " << height_ << std::endl
	    << "gsave % EPS gsave" << std::endl
		<< "2 setlinewidth" << std::endl;

	for(int j=0; j<pathColl.size(); ++j)
	{	
		const BezierPath* bp = pathColl[j];
		for(int i=0; i<bp->GetLength(); ++i)
		{
			for(int m=0; m<4; ++m)
			{
				if( m!=0 || i==0 ) // no start vertex but at the start
				{
					f << bp->GetX(i,m) << " " << bp->GetY(i,m) << " ";
				}
				
				if( i==0 && m==0 ) // moveto only for first vertex of BezierCollection
				{
					f << "moveto " << std::endl;
				}
				if( m==3 )
				{
					f << "curveto" << std::endl;
				}
			}
		}
		f << "stroke" << std::endl;
	}

	f << "grestore %% EPS grestore" << std::endl;
}

void BezierPathCollection::CreateTestBitmap(Bitmap& bm) const
{
	assert(hasPath);

	bm.CreateBlack(width_, height_);

	// draw Bzier curve
	for(int j=0; j<pathColl.size(); ++j)
	{	
		BezierPath* bp = pathColl[j];
		for(int i=0; i<bp->GetLength(); ++i)
		{
			for(double t=0.0; t<=1.0001; t+=0.005)
			{
				int x = (int)(bp->GetXt(i,t)+0.5);
				if( x<0 || x>=width_ ) continue;
				int y = (int)(bp->GetYt(i,t)+0.5);
				if( y<0 || y>=height_ ) continue;

				bm.SetPixel(x, y, 0, 0x80+((59*j+123)&0x7F));
				bm.SetPixel(x, y, 1, 0x80+((41*j+213)&0x7F));
				bm.SetPixel(x, y, 2, 0x80+((83*j+98)&0x7F));
			}
		}
	}

	// draw Bzier points
	for(int j=0; j<pathColl.size(); ++j)
	{	
		BezierPath* bp = pathColl[j];
		for(int i=0; i<bp->GetLength(); ++i)
		{
			for(int k=0; k<4; k++)
			{
				int x = (int)(bp->GetX(i,k)+0.5);
				if( x<0 || x>=width_ ) continue;
				int y = (int)(bp->GetY(i,k)+0.5);
				if( y<0 || y>=height_ ) continue;
				const int vertexColor[][3] = {{255,255,255}, {255,0,0}, {0,255,0}, {255,255,0}};
				bm.SetPixel( x, y, 0, vertexColor[k][0]);
				bm.SetPixel( x, y, 1, vertexColor[k][1]);
				bm.SetPixel( x, y, 2, vertexColor[k][2]);
			}
		}
	}
}
