// (c) 2002 by Jrn Loviscach

#include <assert.h>

#include "pixelpath.h"
#include "adapt.h"

PixelPath::PixelPath()
: hasPath(false)
{}

PixelPath::~PixelPath()
{
}

// returns true if path found, else false
// found path is cleared from BMP
// needs "framed" BMP
PixelPath::Trace(Bitmap& bmp, int channel, double depthDiff)
{
	assert(!hasPath);
	assert(channel == 0 || channel == 1);

	// relative position of the nearest neighbors
	const int xr[] = {1, 1, 1, 0, -1, -1, -1, 0};
	const int yr[] = {1, 0, -1, -1, -1, 0, 1, 1};

	// seek pixels with exactly two nearest neighbors
	// on not too different depths
	int x, y, invD; // later contain the found pixel (if any)
	int direction; // and where we can go
	bool foundInnerPixel = false;

	for(x=1; x<bmp.GetWidth()-1; ++x)
	{
		for(y=1; y<bmp.GetHeight()-1; ++y)
		{
			if(bmp.GetPixel(x, y, channel) == 255)
			{
				invD = bmp.GetPixel(x, y, 2);
				int count = 0;
				for(direction=0; direction<8; direction++)
				{
					int xNew = x + xr[direction];
					int yNew = y + yr[direction];
					int invDNew = bmp.GetPixel(xNew, yNew, 2);

					if(bmp.GetPixel(xNew, yNew, channel) == 255
					   && abs(invDNew-invD) < depthDiff)
					{
						++count;
						if(count == 2)
						{
							foundInnerPixel = true;
							break; // this keeps x, y, invD on their recent values
						}
					}
				}
			}
			if(foundInnerPixel)
			{
				break;
			}
		}
		if(foundInnerPixel)
		{
			break;
		}
	}

	if(!foundInnerPixel)
	{
		return false;
	}

	hasPath = true;
	
	// look along one direction til the end
	// i.e. no new neighbor, all neighbors already visited
	// look first in straight direction
	
	// x, y, direction are initialized already

	// visit neighbors in this order,
	// i.e. straight direction first
	const int growingAngle[] = {0, 1, 7, 2, 6, 3, 4};

	// store the starting point
	pixelQueue.push_back(Pixel(x, y, bmp.GetPixel(x, y, 2)));
	
	// walk path, first forward from found pixel, then backward
	for(int dir=0; dir<2; ++dir) // d == 0: forward, d == 1: backward
	{
		int xNew = x;
		int yNew = y;
		int invDNew = invD;

		// U-turn on second round
		int startDirection = (dir == 0 ? direction : (startDirection + 4)%8);
		
		bool haveNext = true;
		while(haveNext)
		{
			int x1 = xNew;
			int y1 = yNew;
			int invD1 = invDNew;

			int newDirection;
			haveNext = false;
			for(int i=0; i<7; ++i) // 7, because we're already done with one, namely (x,y)
			{
				// first look further along old direction
				newDirection = (growingAngle[i] + direction)%8;
				xNew = x1 + xr[newDirection];
				yNew = y1 + yr[newDirection];
				invDNew = bmp.GetPixel(xNew, yNew, 2);
				if( (bmp.GetPixel(xNew, yNew, channel) == 255)
					&& abs(invDNew-invD1) < depthDiff )
				{
					haveNext = true;
					direction = newDirection;
					break;
				}
			}

			// store found pixel
			if( dir==0 )
			{
				pixelQueue.push_back(Pixel(xNew, yNew, invDNew));
			}
			else // dir==1
			{
				pixelQueue.push_front(Pixel(xNew, yNew, invDNew));
			}
			// clear visited pixel
			bmp.SetPixel(xNew, yNew, channel, 0);
		}
	}

	return true;
}

// remove offset introduced by "framing" and by tracing one side
void PixelPath::Unframe()
{
	assert(hasPath);

	for(int i=0; i<GetLength(); ++i)
	{
		// looks better if we leave y as it is (and would crash above upper range)
		// ++pixelQueue[i].y;
		++pixelQueue[i].x;
	}
}

// try to draw from left to right and from top to bottom
void PixelPath::ConformDirection()
{
	assert(hasPath);

	// this could be simplified using only the initial and final point
	int numRightDown = 0; // counts how often path walks to right or (exactly) down
	for(int i=0; i<GetLength()-1; ++i)
	{
		int dx = pixelQueue[i+1].x - pixelQueue[i].x;
		int dy = pixelQueue[i+1].y - pixelQueue[i].y;
		if( dx > 0 || (dx == 0 && dy == -1) )
		{
			++numRightDown;
		}
	}

	if(numRightDown >= GetLength()/2)
	{
		return; // path direction is ok
	}
	// possible improvement: if above inequality is only weakly fulfilled,
	// additionally favor clockwise movement

	// change direction
	for(int i=0; i<GetLength()/2; ++i)
	{
		int xa = pixelQueue[i].x;
		int ya = pixelQueue[i].y;
		pixelQueue[i].x = pixelQueue[GetLength()-i-1].x;
		pixelQueue[i].y = pixelQueue[GetLength()-i-1].y;
		pixelQueue[GetLength()-i-1].x = xa;
		pixelQueue[GetLength()-i-1].y = ya;
	}
}
